From: Sean Christopherson <seanjc@google.com>
To: Hou Wenlong <houwenlong.hwl@antgroup.com>
Cc: kvm@vger.kernel.org, Lai Jiangshan <jiangshan.ljs@antgroup.com>,
Paolo Bonzini <pbonzini@redhat.com>,
Thomas Gleixner <tglx@linutronix.de>,
Ingo Molnar <mingo@redhat.com>, Borislav Petkov <bp@alien8.de>,
Dave Hansen <dave.hansen@linux.intel.com>,
x86@kernel.org, "H. Peter Anvin" <hpa@zytor.com>,
linux-kernel@vger.kernel.org
Subject: Re: [PATCH 4/7] KVM: x86: Consolidate KVM_GUESTDBG_SINGLESTEP check into the kvm_inject_emulated_db()
Date: Thu, 11 Dec 2025 09:19:39 -0800 [thread overview]
Message-ID: <aTr9Kx9PjLuV9bi1@google.com> (raw)
In-Reply-To: <20251211140520.GC42509@k08j02272.eu95sqa>
On Thu, Dec 11, 2025, Hou Wenlong wrote:
> On Fri, Dec 05, 2025 at 09:58:04AM -0800, Sean Christopherson wrote:
> > But I think the WARN will be subject to false positives. KVM doesn't emulate data
> > #DBs, but it does emulate code #DBs, and fault-like code #DBs can be coincident
> > with trap-like single-step #DBs. Ah, but kvm_vcpu_check_code_breakpoint() doesn't
> > account for RFLAGS.TF. That should probably be addressed in this series, especially
> > since it's consolidating KVM_GUESTDBG_SINGLESTEP handling.
>
> Sorry, I didn't follow it, how fault-like code #DBs can be coincident
> with trap-like single-step #DBs, could you provide an example?
Ya, here's a KUT testcase that applies on top of
https://lore.kernel.org/all/20251126191736.907963-1-seanjc@google.com.
---
x86/debug.c | 43 +++++++++++++++++++++++++++++++++++++++----
1 file changed, 39 insertions(+), 4 deletions(-)
diff --git a/x86/debug.c b/x86/debug.c
index 8177575c..313d854e 100644
--- a/x86/debug.c
+++ b/x86/debug.c
@@ -92,6 +92,7 @@ typedef unsigned long (*db_test_fn)(void);
typedef void (*db_report_fn)(unsigned long, const char *);
static unsigned long singlestep_with_movss_blocking_and_dr7_gd(void);
+static unsigned long singlestep_with_code_db(void);
static unsigned long singlestep_with_sti_hlt(void);
static void __run_single_step_db_test(db_test_fn test, db_report_fn report_fn)
@@ -106,11 +107,12 @@ static void __run_single_step_db_test(db_test_fn test, db_report_fn report_fn)
report_fn(start, "");
/*
* MOV DR #GPs at CPL>0, don't try to run the DR7.GD test in usermode.
* Likewise for HLT.
*/
- if (test == singlestep_with_movss_blocking_and_dr7_gd
- || test == singlestep_with_sti_hlt)
+ if (test == singlestep_with_movss_blocking_and_dr7_gd ||
+ test == singlestep_with_code_db ||
+ test == singlestep_with_sti_hlt)
return;
n = 0;
@@ -163,6 +165,38 @@ static noinline unsigned long singlestep_basic(void)
return start;
}
+static void report_singlestep_with_code_db(unsigned long start, const char *usermode)
+{
+ report(n == 3 &&
+ dr6[0] == (DR6_ACTIVE_LOW | DR6_BS | DR6_TRAP2) && db_addr[0] == start &&
+ is_single_step_db(dr6[1]) && db_addr[1] == start + 1 &&
+ is_single_step_db(dr6[2]) && db_addr[2] == start + 1 + 1,
+ "%sSingle-step + code #DB test", usermode);
+}
+
+static noinline unsigned long singlestep_with_code_db(void)
+{
+ unsigned long start;
+
+ asm volatile (
+ "lea 1f(%%rip), %0\n\t"
+ "mov %0, %%dr2\n\t"
+ "mov $" xstr(DR7_FIXED_1 | DR7_EXECUTE_DRx(2) | DR7_GLOBAL_ENABLE_DR2) ", %0\n\t"
+ "mov %0, %%dr7\n\t"
+ "pushf\n\t"
+ "pop %%rax\n\t"
+ "or $(1<<8),%%rax\n\t"
+ "push %%rax\n\t"
+ "popf\n\t"
+ "and $~(1<<8),%%rax\n\t"
+ "1:push %%rax\n\t"
+ "popf\n\t"
+ "lea 1b(%%rip), %0\n\t"
+ : "=r" (start) : : "rax"
+ );
+ return start;
+}
+
static void report_singlestep_emulated_instructions(unsigned long start,
const char *usermode)
{
@@ -517,6 +551,7 @@ int main(int ac, char **av)
n, db_addr[0], dr6[0]);
run_ss_db_test(singlestep_basic);
+ run_ss_db_test(singlestep_with_code_db);
run_ss_db_test(singlestep_emulated_instructions);
run_ss_db_test(singlestep_with_sti_blocking);
run_ss_db_test(singlestep_with_movss_blocking);
@@ -106,11 +107,12 @@ static void __run_single_step_db_test(db_test_fn test, db_report_fn report_fn)
report_fn(start, "");
/*
- * MOV DR #GPs at CPL>0, don't try to run the DR7.GD test in usermode.
- * Likewise for HLT.
+ * MOV DR #GPs at CPL>0, don't try to run the DR7.GD or code #DB tests
+ * in usermode. Likewise for HLT.
*/
base-commit: 23071a886edbe303fb964c5c386750b0b458dbfb
--
next prev parent reply other threads:[~2025-12-11 17:19 UTC|newest]
Thread overview: 22+ messages / expand[flat|nested] mbox.gz Atom feed top
2025-09-10 2:49 [PATCH 0/7] KVM: x86: Improve the handling of debug exceptions during instruction emulation Hou Wenlong
2025-09-10 2:49 ` [PATCH 1/7] KVM: x86: Set guest DR6 by kvm_queue_exception_p() in " Hou Wenlong
2025-09-10 2:49 ` [PATCH 2/7] KVM: x86: Check guest debug in DR access " Hou Wenlong
2025-12-05 17:51 ` Sean Christopherson
2025-09-10 2:49 ` [PATCH 3/7] KVM: x86: Only check effective code breakpoint in emulation Hou Wenlong
2025-09-10 2:49 ` [PATCH 4/7] KVM: x86: Consolidate KVM_GUESTDBG_SINGLESTEP check into the kvm_inject_emulated_db() Hou Wenlong
2025-12-05 17:58 ` Sean Christopherson
2025-12-11 14:05 ` Hou Wenlong
2025-12-11 17:19 ` Sean Christopherson [this message]
2025-12-12 9:46 ` Hou Wenlong
2025-12-12 17:53 ` Sean Christopherson
2025-12-13 16:15 ` Hou Wenlong
2025-12-17 0:43 ` Sean Christopherson
2025-09-10 2:49 ` [PATCH 5/7] KVM: VMX: Set 'BS' bit in pending debug exceptions during instruction emulation Hou Wenlong
2025-12-05 18:20 ` Sean Christopherson
2025-12-11 14:01 ` Hou Wenlong
2025-09-10 2:49 ` [PATCH 6/7] KVM: selftests: Verify guest debug DR7.GD checking " Hou Wenlong
2025-12-05 18:21 ` Sean Christopherson
2025-09-10 2:49 ` [PATCH 7/7] KVM: selftests: Verify 'BS' bit checking in pending debug exception during VM entry Hou Wenlong
2025-12-05 18:23 ` Sean Christopherson
2025-12-11 13:21 ` Hou Wenlong
2025-12-18 13:40 ` Hou Wenlong
Reply instructions:
You may reply publicly to this message via plain-text email
using any one of the following methods:
* Save the following mbox file, import it into your mail client,
and reply-to-all from there: mbox
Avoid top-posting and favor interleaved quoting:
https://en.wikipedia.org/wiki/Posting_style#Interleaved_style
* Reply using the --to, --cc, and --in-reply-to
switches of git-send-email(1):
git send-email \
--in-reply-to=aTr9Kx9PjLuV9bi1@google.com \
--to=seanjc@google.com \
--cc=bp@alien8.de \
--cc=dave.hansen@linux.intel.com \
--cc=houwenlong.hwl@antgroup.com \
--cc=hpa@zytor.com \
--cc=jiangshan.ljs@antgroup.com \
--cc=kvm@vger.kernel.org \
--cc=linux-kernel@vger.kernel.org \
--cc=mingo@redhat.com \
--cc=pbonzini@redhat.com \
--cc=tglx@linutronix.de \
--cc=x86@kernel.org \
/path/to/YOUR_REPLY
https://kernel.org/pub/software/scm/git/docs/git-send-email.html
* If your mail client supports setting the In-Reply-To header
via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line
before the message body.
This is an external index of several public inboxes,
see mirroring instructions on how to clone and mirror
all data and code used by this external index.